# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric library for Let's Encrypt certbot
"""

from __future__ import unicode_literals, absolute_import

from fabric.api import sudo, cd, abort
from fabric.contrib.files import exists

from rattail.fablib import apt, get_debian_version


def install():
    """
    Install the Let's Encrypt certbot utility
    """
    version = get_debian_version()

    # debian 7 wheezy
    if 7 <= version < 8:
        if not exists('/usr/local/src/certbot'):
            with cd('/usr/local/src'):
                sudo('git clone https://github.com/certbot/certbot')
        sudo('ln --symbolic --force /usr/local/src/certbot/certbot-auto /usr/local/bin/certbot')

    # debian 8 jessie
    elif 8 <= version < 9:
        apt.add_source('deb http://ftp.debian.org/debian jessie-backports main')
        apt.install('python-certbot-apache', target_release='jessie-backports')

    # other..? will have to investigate when this comes up
    else:
        abort("don't know how to install certbot on debian version {}".format(version))


def certonly(*domains):
    """
    Obtain SSL certificates
    """
    domains = ['--domain {}'.format(domain) for domain in domains]
    sudo('certbot certonly --noninteractive --apache {}'.format(' '.join(domains)))



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

from sqlalchemy.exc import IntegrityError

from rattail.db import model
# from rattail.db.changes import record_changes
from rattail.tests import DataTestCase


class SAErrorHelper(object):

    def integrity_or_flush_error(self):
        try:
            from sqlalchemy.exc import FlushError
        except ImportError:
            return IntegrityError
        else:
            return (IntegrityError, FlushError)


class TestCustomerPerson(DataTestCase, SAErrorHelper):

    def test_repr(self):
        assoc = model.CustomerPerson(uuid='whatever')
        self.assertEqual(repr(assoc), "CustomerPerson(uuid=u'whatever')")

    def test_customer_required(self):
        assoc = model.CustomerPerson(person=model.Person())
        self.session.add(assoc)
        self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
        assoc.customer = model.Customer()
        self.session.add(assoc)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

    def test_person_required(self):
        assoc = model.CustomerPerson(customer=model.Customer())
        self.session.add(assoc)
        self.assertRaises(IntegrityError, self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
        assoc.person = model.Person()
        self.session.add(assoc)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

    def test_ordinal_autoincrement(self):
        customer = model.Customer()
        self.session.add(customer)
        assoc = model.CustomerPerson(person=model.Person())
        customer._people.append(assoc)
        self.session.commit()
        self.assertEqual(assoc.ordinal, 1)
        assoc = model.CustomerPerson(person=model.Person())
        customer._people.append(assoc)
        self.session.commit()
        self.assertEqual(assoc.ordinal, 2)


class TestCustomerGroupAssignment(DataTestCase, SAErrorHelper):

    def test_repr(self):
        assignment = model.CustomerGroupAssignment(uuid='whatever')
        self.assertEqual(repr(assignment), "CustomerGroupAssignment(uuid=u'whatever')")

    def test_customer_required(self):
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        self.session.add(assignment)
        self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
        assignment.customer = model.Customer()
        self.session.add(assignment)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

    def test_group_required(self):
        assignment = model.CustomerGroupAssignment(customer=model.Customer())
        self.session.add(assignment)
        self.assertRaises(IntegrityError, self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
        assignment.group = model.CustomerGroup()
        self.session.add(assignment)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

    def test_ordinal_autoincrement(self):
        customer = model.Customer()
        self.session.add(customer)
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        customer._groups.append(assignment)
        self.session.commit()
        self.assertEqual(assignment.ordinal, 1)
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        customer._groups.append(assignment)
        self.session.commit()
        self.assertEqual(assignment.ordinal, 2)


# class TestCustomerEmailAddress(DataTestCase):

#     def test_pop(self):
#         customer = model.Customer()
#         customer.add_email_address('fred.home@mailinator.com')
#         customer.add_email_address('fred.work@mailinator.com')
#         self.session.add(customer)
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

#         while customer.emails:
#             customer.emails.pop()
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

#         # changes weren't being recorded
#         self.assertEqual(self.session.query(model.Change).count(), 0)

#     def test_pop_with_changes(self):
#         record_changes(self.session)

#         customer = model.Customer()
#         customer.add_email_address('fred.home@mailinator.com')
#         customer.add_email_address('fred.work@mailinator.com')
#         self.session.add(customer)
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

#         while customer.emails:
#             customer.emails.pop()
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

#         # changes should have been recorded
#         changes = self.session.query(model.Change)
#         self.assertEqual(changes.count(), 3)

#         customer_change = changes.filter_by(class_name='Customer').one()
#         self.assertEqual(customer_change.uuid, customer.uuid)
#         self.assertFalse(customer_change.deleted)

#         email_changes = changes.filter_by(class_name='CustomerEmailAddress')
#         self.assertEqual(email_changes.count(), 2)
#         self.assertEqual([x.deleted for x in email_changes], [True, True])


class TestLabelProfile(DataTestCase):

    def test_get_printer_setting(self):
        profile = model.LabelProfile()
        self.session.add(profile)

        self.assertTrue(profile.uuid is None)
        setting = profile.get_printer_setting('some_setting')
        self.assertTrue(setting is None)
        self.assertTrue(profile.uuid is None)

        profile.uuid = 'some_uuid'
        self.session.add(model.Setting(
                name='labels.some_uuid.printer.some_setting',
                value='some_value'))
        self.session.flush()
        setting = profile.get_printer_setting('some_setting')
        self.assertEquals(setting, 'some_value')

    def test_save_printer_setting(self):
        self.assertEqual(self.session.query(model.Setting).count(), 0)
        profile = model.LabelProfile()
        self.session.add(profile)

        self.assertTrue(profile.uuid is None)
        profile.save_printer_setting('some_setting', 'some_value')
        self.assertFalse(profile.uuid is None)
        self.assertEqual(self.session.query(model.Setting).count(), 1)

        profile.uuid = 'some_uuid'
        profile.save_printer_setting('some_setting', 'some_value')
        self.assertEqual(self.session.query(model.Setting).count(), 2)
        setting = self.session.query(model.Setting)\
            .filter_by(name='labels.some_uuid.printer.some_setting')\
            .one()
        self.assertEqual(setting.value, 'some_value')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Upgrade handlers
"""

from __future__ import unicode_literals, absolute_import

import os
import shutil
import subprocess

from rattail.time import make_utc
from rattail.util import load_object


class UpgradeHandler(object):
    """
    Base class and default implementation for upgrade handlers.
    """

    def __init__(self, config):
        self.config = config
        self.enum = config.get_enum()

    def executable(self, upgrade):
        """
        This method should return a boolean indicating whether or not execution
        should be allowed for the upgrade, given its current condition.  The
        default simply returns ``True`` unless the upgrade has already been
        executed.
        """
        if upgrade is None:
            return True
        return not bool(upgrade.executed)

    def mark_executing(self, upgrade):
        upgrade.executing = True
        upgrade.status_code = self.enum.UPGRADE_STATUS_EXECUTING

    def do_execute(self, upgrade, user, **kwargs):
        result = self.execute(upgrade, user, **kwargs)
        upgrade.executing = False
        if result:
            upgrade.status_code = self.enum.UPGRADE_STATUS_SUCCEEDED
        else:
            upgrade.status_code = self.enum.UPGRADE_STATUS_FAILED
        upgrade.executed = make_utc()
        upgrade.executed_by = user

    def execute(self, upgrade, user, progress=None, **kwargs):
        """
        Execute the given upgrade, as the given user.
        """
        before_path = self.config.upgrade_filepath(upgrade.uuid, filename='requirements.before.txt', makedirs=True)
        self.record_requirements_snapshot(before_path)

        stdout_path = self.config.upgrade_filepath(upgrade.uuid, filename='stdout.log')
        stderr_path = self.config.upgrade_filepath(upgrade.uuid, filename='stderr.log')
        cmd = self.config.upgrade_command()
        with open(stdout_path, 'wb') as stdout:
            with open(stderr_path, 'wb') as stderr:
                upgrade.exit_code = subprocess.call(cmd, stdout=stdout, stderr=stderr)

        after_path = self.config.upgrade_filepath(upgrade.uuid, filename='requirements.after.txt')
        self.record_requirements_snapshot(after_path)
        return upgrade.exit_code == 0

    def record_requirements_snapshot(self, path):
        pip = self.get_pip_path()
        logpath = os.path.join(self.config.workdir(), 'pip.log')
        with open(path, 'wb') as stdout:
            subprocess.call([pip, '--log', logpath, 'freeze'], stdout=stdout)

    def get_pip_path(self):
        path = os.path.join(self.config.appdir(), os.pardir, 'bin', 'pip')
        return os.path.abspath(path)

    def delete_files(self, upgrade):
        """
        Delete all data files for the given upgrade.
        """
        path = self.config.upgrade_filepath(upgrade.uuid)
        if os.path.exists(path):
            shutil.rmtree(path)


def get_upgrade_handler(config, default=None):
    """
    Returns an upgrade handler object.
    """
    spec = config.get('rattail.upgrades', 'handler', default=default)
    if spec:
        handler = load_object(spec)(config)
    else:
        handler = UpgradeHandler(config)
    return handler



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Windows User Utilities
"""

from __future__ import unicode_literals

import socket
import logging


log = logging.getLogger(__name__)


def user_exists(username, server=None):
    """
    Determine if a Windows user account exists.

    :param username: Username to be checked.

    :param server: Optional server name.  If not specified, only local accounts
       will be checked.
    """
    import win32net
    from pywintypes import error

    # This constant doesn't seem to be importable from anywhere.
    NERR_UserNotFound = 2221

    try:
        info = win32net.NetUserGetInfo(server, username, 0)
    except error, e:
        if e.winerror == NERR_UserNotFound:
            return False
        raise
    return True


def create_user(username, password, full_name=None, comment=None):
    """
    Create a system user account for Rattail.
    """
    import win32net
    import win32netcon

    if user_exists(username):
        log.warning("user already exists: {0}".format(username))
        return False

    if not full_name:
        full_name = "{0} User".format(username.capitalize())
    if not comment:
        comment = "System user account for Rattail applications"

    win32net.NetUserAdd(None, 2, {
            'name': username,
            'password': password,
            'priv': win32netcon.USER_PRIV_USER,
            'comment': comment,
            'flags': (win32netcon.UF_NORMAL_ACCOUNT
                      | win32netcon.UF_PASSWD_CANT_CHANGE
                      | win32netcon.UF_DONT_EXPIRE_PASSWD),
            'full_name': full_name,
            'acct_expires': win32netcon.TIMEQ_FOREVER,
            })

    win32net.NetLocalGroupAddMembers(None, 'Users', 3, [
            {'domainandname': r'{0}\{1}'.format(socket.gethostname(), username)}])

    hide_user_account(username)
    return True


def hide_user_account(username):
    """
    Hide a user account from the Welcome screen.

    This also hides it from the User Accounts control panel applet.
    """

    import win32api
    import win32con
    from rattail.win32 import is_64bit

    if is_64bit():
        key64 = win32api.RegOpenKeyEx(
            win32con.HKEY_LOCAL_MACHINE,
            r'SOFTWARE\Microsoft\Windows NT\CurrentVersion\Winlogon',
            0, win32con.KEY_ALL_ACCESS|win32con.KEY_WOW64_64KEY)
        key = win32api.RegCreateKey(key64, 'SpecialAccounts\\UserList')
        win32api.RegCloseKey(key64)
    else:
        key = win32api.RegCreateKey(
            win32con.HKEY_LOCAL_MACHINE,
            'SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon\\SpecialAccounts\\UserList')

    win32api.RegSetValueEx(key, username, 0, win32con.REG_DWORD, 0)
    win32api.RegCloseKey(key)


def allow_logon_as_service(username):
    """
    Grant the "Log On As Service" right to a local user account.

    .. note::
       The user account is assumed to already exist.
    """
    import win32security
    import win32con

    sid, domain, sid_type = win32security.LookupAccountName(None, username)
    policy = win32security.LsaOpenPolicy(None, win32con.GENERIC_WRITE)
    win32security.LsaAddAccountRights(policy, sid, [win32security.SE_SERVICE_LOGON_NAME])
    win32security.LsaClose(policy)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime
from unittest import TestCase

import pytz

from rattail import time
from rattail.config import make_config
from rattail.exceptions import ConfigurationError


class TestLocaltime(TestCase):

    def setUp(self):
        self.config = make_config([])
        self.config.set('rattail', 'timezone.default', 'America/Los_Angeles')
        self.config.set('rattail', 'timezone.missouri', 'America/Chicago')

    def test_default_returns_current_time_in_default_timezone(self):
        zone = pytz.timezone('America/Los_Angeles')
        before = zone.normalize(pytz.utc.localize(datetime.datetime.utcnow()).astimezone(zone))
        local = time.localtime(self.config)
        after = zone.normalize(pytz.utc.localize(datetime.datetime.utcnow()).astimezone(zone))
        self.assertTrue(before <= local)
        self.assertTrue(local <= after)
        self.assertEqual(local.tzinfo.zone, 'America/Los_Angeles')

    def test_provided_time_is_localized_to_default_timezone(self):
        t = datetime.datetime(2014, 7, 26, 8, 28, 41)
        t = time.localtime(self.config, t)
        self.assertEqual(t.year, 2014)
        self.assertEqual(t.month, 7)
        self.assertEqual(t.day, 26)
        self.assertEqual(t.hour, 8)
        self.assertEqual(t.minute, 28)
        self.assertEqual(t.second, 41)
        self.assertEqual(t.tzinfo.zone, 'America/Los_Angeles')

    def test_provided_time_is_localized_to_custom_timezone(self):
        t = datetime.datetime(2014, 7, 26, 8, 28, 41)
        t = time.localtime(self.config, t, 'missouri')
        self.assertEqual(t.year, 2014)
        self.assertEqual(t.month, 7)
        self.assertEqual(t.day, 26)
        self.assertEqual(t.hour, 8)
        self.assertEqual(t.minute, 28)
        self.assertEqual(t.second, 41)
        self.assertEqual(t.tzinfo.zone, 'America/Chicago')


class TestTimezone(TestCase):

    def setUp(self):
        self.config = make_config([])

    def test_default_timezone_returned_by_default(self):
        self.config.set('rattail', 'timezone.default', 'Antarctica/Rothera')
        zone = time.timezone(self.config)
        self.assertEqual(zone.zone, 'Antarctica/Rothera')

    def test_local_timezone_returned_if_no_default(self):
        self.config.set('rattail', 'timezone.local', 'Africa/Lagos')
        zone = time.timezone(self.config)
        self.assertEqual(zone.zone, 'Africa/Lagos')

    def test_custom_timezone_returned_instead_of_default(self):
        self.config.set('rattail', 'timezone.another', 'Canada/Yukon')
        zone = time.timezone(self.config, key='another')
        self.assertEqual(zone.zone, 'Canada/Yukon')

    def test_config_error_if_no_timezone_defined(self):
        self.assertRaises(ConfigurationError, time.timezone, self.config)



# -*- coding: utf-8 -*-
"""add upgrade.status

Revision ID: 139fea9de0a0
Revises: e5c3eab67b28
Create Date: 2017-08-07 19:32:43.142842

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '139fea9de0a0'
down_revision = u'e5c3eab67b28'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # upgrade
    op.add_column('upgrade', sa.Column('exit_code', sa.Integer(), nullable=True))
    op.add_column('upgrade', sa.Column('status_code', sa.Integer(), nullable=True))


def downgrade():

    # upgrade
    op.drop_column('upgrade', 'status_code')
    op.drop_column('upgrade', 'exit_code')



# -*- coding: utf-8 -*-
"""make batch filename nullable

Revision ID: 36ffa8b83782
Revises: c384c65995db
Create Date: 2016-11-16 00:58:42.069307

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '36ffa8b83782'
down_revision = u'c384c65995db'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # batch_handheld
    op.alter_column('batch_handheld', 'filename',
               existing_type=sa.VARCHAR(length=255),
               nullable=True)


def downgrade():

    # batch_handheld
    op.alter_column('batch_handheld', 'filename',
               existing_type=sa.VARCHAR(length=255),
               nullable=False)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Palm OS Application Interface
"""

from __future__ import absolute_import

import os
import os.path
import socket
import getpass
import datetime
import logging

from rattail.config import make_config
from rattail.csvutil import DictWriter

try:
    import pythoncom
except ImportError:
    # Mock out for Linux.
    class Object(object):
        pass
    pythoncom = Object()
    pythoncom.CLSCTX_LOCAL_SERVER = 4


log = logging.getLogger(__name__)


class PalmConduit(object):
    """
    Implements a conduit for Palm's HotSync Manager.
    """

    _reg_clsid_ = '{F2FDDEEC-254F-42C3-8801-C41E8A243F13}'
    _reg_progid_ = 'Rattail.PalmConduit'
    _reg_desc_ = "Rattail Conduit for Palm HotSync Manager"

    # Don't let pythoncom guess this, for several reasons.  This way Python
    # will go about determining its path etc. as normal, and __name__ will be
    # "rattail.palm" instead of just "palm".
    _reg_class_spec_ = 'rattail.palm.PalmConduit'

    # Don't let HotSync Manager run our code in-process, so that we may launch
    # wxPython dialogs as needed for configuration etc.
    _reg_clsctx_ = pythoncom.CLSCTX_LOCAL_SERVER

    _typelib_guid_ = '{6FD7A7A0-FA1F-11D2-AC32-006008E3F0A2}'
    _typelib_version_ = 1, 0
    _com_interfaces_ = ['IPDClientNotify']

    _public_methods_ = ['BeginProcess', 'CfgConduit',
                        'ConfigureConduit', 'GetConduitInfo']

    def BeginProcess(self):
        """
        Called by HotSync Manager when synchronization is ready to happen.
        This method implements the actual data sync.
        """

        from win32com.client import Dispatch

        config = make_config()
        data_dir = config.require('rattail.palm', 'collection_dir')
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)

        db_query = Dispatch('PDDirect.PDDatabaseQuery')
        db = db_query.OpenRecordDatabase('Rattail_Scan', 'PDDirect.PDRecordAdapter')
        if db.RecordCount:
            
            sys_adapter = Dispatch('PDDirect.PDSystemAdapter')
            fname = '%s,%s,%s,%s.csv' % (socket.gethostname(), getpass.getuser(),
                                         sys_adapter.PDUserInfo.UserName,
                                         datetime.datetime.now().strftime('%Y-%m-%d %H-%M-%S'))
            data_path = os.path.join(data_dir, fname)
            log.info("PalmConduit.BeginProcess: writing %u handheld records to file: %s" %
                     (db.RecordCount, data_path))

            data_file = open(data_path, 'wb')
            writer = DictWriter(data_file, ['upc', 'cases', 'units'])
            writer.writeheader()
            for i in range(db.RecordCount):
                rec, unique_id, category, attrs = db.ReadByIndex(i)

                writer.writerow({
                        'upc': rec[:15].rstrip('\x00'),
                        'cases': rec[15:19].rstrip('\x00'),
                        'units': rec[19:23].rstrip('\x00'),
                        })

            data_file.close()

            log.info("PalmConduit.BeginProcess: removing all records from handheld")
            db.RemoveSet(0)
            log.info("PalmConduit.BeginProcess: done")

        else:
            log.info("PalmConduit.BeginProcess: nothing to do")

        return False

    def CfgConduit(self, nCreatorId, nUserId, bstrUserName, bstrPathName,
                   nSyncPerm, nSyncTemp, nSyncNew, nSyncPref):
        pass

    def ConfigureConduit(self, pPathName, pRegistry, nSyncPref, nSyncType):
        pass

    def GetConduitInfo(self, infoType, dwCreatorID, dwUserID, bstrUsername):
        return None


def get_conduit_manager():
    """
    Load an instance of the Palm Conduit Manager.
    """

    from win32com.client import Dispatch
    from pywintypes import com_error
    from winerror import CO_E_CLASSSTRING
    from rattail.exceptions import PalmConduitManagerNotFound

    try:
        conduit_mgr = Dispatch('PDStandard.PDSystemCondMgr')
    except com_error, error:
        if error.hresult == CO_E_CLASSSTRING: # "Invalid class string"
            raise PalmConduitManagerNotFound()
        raise

    return conduit_mgr


def get_creator_id(conduit_mgr=None):
    """
    Get the "creator ID" for the Rattail Palm conduit.
    """

    if conduit_mgr is None:
        conduit_mgr = get_conduit_manager()
    return conduit_mgr.StringToCreatorID('RTTL')


def conduit_is_registered(conduit_mgr=None):
    """
    Check if the Rattail Palm conduit is registered with HotSync Manager.
    """

    if conduit_mgr is None:
        conduit_mgr = get_conduit_manager()
    rattail_id = get_creator_id(conduit_mgr)
    for creator_id in conduit_mgr.GetConduitList():
        if creator_id == rattail_id:
            return True
    return False


def register_com_server():
    """
    Registers the COM server class which implements the Rattail Palm conduit.
    """

    from win32com.client.gencache import EnsureModule
    from win32com.server.register import RegisterClasses
    from rattail.exceptions import PalmClassicDatabaseTypelibNotFound

    # First we must make sure the Palm Classic Database type library is
    # available to Python.  Note that we specify version 1.0 here.  I'm not
    # sure if that matters or not, but 1.0 was the only registry key found
    # below the type library GUID when I last checked:
    #
    #    HKEY_CLASSES_ROOT\TypeLib\{6FD7A7A0-FA1F-11D2-AC32-006008E3F0A2}
    #
    # TODO: Unfortunately I don't even remember why (or indeed, if) this
    # EnsureModule business is necessary.  If I ever (re)discover that, I
    # should make mention of it here...
    if not EnsureModule('{6FD7A7A0-FA1F-11D2-AC32-006008E3F0A2}', 0, 1, 0):
        raise PalmClassicDatabaseTypelibNotFound()

    RegisterClasses(PalmConduit, quiet=True)


def register_conduit():
    """
    Registers the Rattail Palm conduit with HotSync Manager.
    """

    from win32com.client import Dispatch
    from rattail.exceptions import PalmConduitAlreadyRegistered

    register_com_server()

    conduit_mgr = get_conduit_manager()
    if conduit_is_registered(conduit_mgr):
        raise PalmConduitAlreadyRegistered()

    info = Dispatch('PDStandard.PDConduitInfo')
    info.COMClassID = 'Rattail.PalmConduit'
    info.CreatorID = get_creator_id(conduit_mgr)
    info.DesktopDataDirectory = 'Rattail'
    info.DisplayName = 'Rattail'
    conduit_mgr.RegisterConduit(info)


def unregister_conduit():
    """
    Unregisters the Rattail Palm conduit from HotSync Manager.
    """

    from rattail.exceptions import PalmConduitNotRegistered

    conduit_mgr = get_conduit_manager()
    if not conduit_is_registered(conduit_mgr):
        raise PalmConduitNotRegistered()

    rattail_id = get_creator_id(conduit_mgr)
    conduit_mgr.UnregisterConduit(rattail_id)

    unregister_com_server()


def unregister_com_server():
    """
    Unregisters the COM server class which implements the Rattail Palm conduit.
    """

    from win32com.client import Dispatch
    from pywintypes import com_error
    from winerror import CO_E_CLASSSTRING
    from rattail.exceptions import PalmConduitNotRegistered

    try:
        conduit = Dispatch('Rattail.PalmConduit')
    except com_error, error:
        if error.hresult == CO_E_CLASSSTRING: # "Invalid class string"
            raise PalmConduitNotRegistered()
        raise

    from win32com.server.register import UnregisterClasses
    UnregisterClasses(PalmConduit, quiet=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Excel utilities
"""

from __future__ import unicode_literals, absolute_import

import xlrd

from rattail.util import progress_loop


class ExcelReader(object):
    """
    Basic class for reading Excel files.
    """

    def __init__(self, path, sheet=0, header=0):
        """
        Constructor; opens an Excel file for reading.
        """
        self.book = xlrd.open_workbook(path)
        self.sheet = self.book.sheet_by_index(sheet)
        self.header = header
        self.fields = self.sheet.row_values(self.header)

    def read_rows(self, progress=None):
        rows = []

        def append(row, i):
            values = self.sheet.row_values(row)
            data = dict([(self.fields[j], value)
                         for j, value in enumerate(values)])
            rows.append(data)

        progress_loop(append, range(self.header + 1, self.sheet.nrows), progress,
                      message="Reading data from Excel file")
        return rows



# -*- coding: utf-8 -*-
"""fix user_x_role null bug

Revision ID: 3a7029aed67
Revises: 35b1da3be8f5
Create Date: 2015-08-11 11:36:41.908780

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3a7029aed67'
down_revision = u'35b1da3be8f5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # user_x_role

    # first remove all records with nulls
    user_x_role = sa.sql.table('user_x_role',
                               sa.sql.column('role_uuid'),
                               sa.sql.column('user_uuid'))
    op.execute(user_x_role.delete().where(user_x_role.c.role_uuid == None))
    op.execute(user_x_role.delete().where(user_x_role.c.user_uuid == None))

    # okay, now can disallow nulls
    op.alter_column('user_x_role', 'role_uuid',
               existing_type=sa.String(length=32),
               nullable=False)
    op.alter_column('user_x_role', 'user_uuid',
               existing_type=sa.String(length=32),
               nullable=False)


def downgrade():

    # user_x_role
    op.alter_column('user_x_role', 'user_uuid',
               existing_type=sa.String(length=32),
               nullable=True)
    op.alter_column('user_x_role', 'role_uuid',
               existing_type=sa.String(length=32),
               nullable=True)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data model for pricing batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import Base, BatchMixin, ProductBatchRowMixin, Vendor


class PricingBatch(BatchMixin, Base):
    """
    Primary data model for pricing batches.
    """
    __tablename__ = 'batch_pricing'
    __batchrow_class__ = 'PricingBatchRow'
    batch_key = 'pricing'
    model_title = "Pricing Batch"

    min_diff_threshold = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)


class PricingBatchRow(ProductBatchRowMixin, Base):
    """
    Row of data within a pricing batch.
    """
    __tablename__ = 'batch_pricing_row'
    __batch_class__ = PricingBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='batch_pricing_row_fk_vendor'),
        )

    STATUS_PRICE_UNCHANGED              = 1
    STATUS_PRICE_INCREASE               = 2
    STATUS_PRICE_DECREASE               = 3
    STATUS_PRODUCT_NOT_FOUND            = 4
    STATUS_PRODUCT_MANUALLY_PRICED      = 5
    STATUS_CANNOT_CALCULATE_PRICE       = 6
    STATUS_PRICE_INCREASE_MINOR         = 7
    STATUS_PRICE_DECREASE_MINOR         = 8

    STATUS = {
        STATUS_PRICE_UNCHANGED          : "price not changed",
        STATUS_PRICE_INCREASE           : "price increase",
        STATUS_PRICE_INCREASE_MINOR     : "price increase (minor)",
        STATUS_PRICE_DECREASE           : "price drop",
        STATUS_PRICE_DECREASE_MINOR     : "price drop (minor)",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
        STATUS_PRODUCT_MANUALLY_PRICED  : "product is priced manually",
        STATUS_CANNOT_CALCULATE_PRICE   : "cannot calculate price",
    }

    vendor_uuid = sa.Column(sa.String(length=32), nullable=True)
    vendor = orm.relationship(Vendor)

    regular_unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True)
    discounted_unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True)

    old_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True)
    price_margin = sa.Column(sa.Numeric(precision=5, scale=3), nullable=True)
    price_markup = sa.Column(sa.Numeric(precision=4, scale=2), nullable=True)
    new_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True)
    price_diff = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
PostgreSQL data importers
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import logging

import six

from rattail.importing import BulkImporter, ToSQLAlchemy
from rattail.time import make_utc


log = logging.getLogger(__name__)


class BulkToPostgreSQL(BulkImporter, ToSQLAlchemy):
    """
    Base class for bulk data importers which target PostgreSQL on the local side.
    """

    @property
    def data_path(self):
        return os.path.join(self.config.workdir(),
                            'import_bulk_postgresql_{}.csv'.format(self.model_name))

    def setup(self):
        self.data_buffer = open(self.data_path, 'wb')

    def teardown(self):
        self.data_buffer.close()
        os.remove(self.data_path)
        self.data_buffer = None

    def create_object(self, key, data):
        data = self.prep_data_for_postgres(data)
        self.data_buffer.write('{}\n'.format('\t'.join([data[field] for field in self.fields])).encode('utf-8'))

    def prep_data_for_postgres(self, data):
        data = dict(data)
        for key, value in data.iteritems():
            data[key] = self.prep_value_for_postgres(value)
        return data

    def prep_value_for_postgres(self, value):
        if value is None:
            return '\\N'
        if value is True:
            return 't'
        if value is False:
            return 'f'

        if isinstance(value, datetime.datetime):
            value = make_utc(value, tzinfo=False)
        elif isinstance(value, basestring):
            value = value.replace('\\', '\\\\')
            value = value.replace('\r', '\\r')
            value = value.replace('\n', '\\n')
            value = value.replace('\t', '\\t') # TODO: add test for this

        return six.text_type(value)

    def flush_create_update(self):
        pass

    def flush_create_update_final(self):
        log.info("copying {} data from buffer to PostgreSQL".format(self.model_name))
        self.data_buffer.close()
        self.data_buffer = open(self.data_path, 'rb')
        cursor = self.session.connection().connection.cursor()
        table_name = '"{}"'.format(self.model_table.name)
        cursor.copy_from(self.data_buffer, table_name, columns=self.fields)
        log.debug("PostgreSQL data copy completed")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Fabric Library
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import tempfile

from fabric.api import sudo, run, cd, local, env, settings, put as fab_put
from fabric.contrib.files import exists, append, upload_template as fab_upload_template, is_link

from mako.template import Template


def get_debian_version():
    """
    Fetch the version of Debian running on the target system.
    """
    version = run('cat /etc/debian_version')
    match = re.match(r'^(\d+\.\d+)$', version)
    if match:
        return float(match.group(1))


def get_ubuntu_version():
    """
    Fetch the version of Ubuntu running on the target system
    """
    info = run('cat /etc/lsb-release')
    match = re.search(r'DISTRIB_RELEASE=(\d+\.\d+)', info)
    if match:
        return float(match.group(1))


def mkdir(paths, owner='root:root', mode=None):
    """
    Recursively make one or more directories.
    """
    if isinstance(paths, basestring):
        paths = [paths]
    sudo('mkdir --parents {0}'.format(' '.join(paths)))
    if owner != 'root:root':
        if ':' not in owner:
            owner = '{0}:{0}'.format(owner)
        sudo('chown {} {}'.format(owner, ' '.join(paths)))
    if mode is not None:
        sudo('chmod {0} {1}'.format(mode, ' '.join(paths)))


def make_system_user(name='rattail', home='/srv/rattail', uid=None, shell=None):
    """
    Make a new system user account, with the given home folder and shell path.
    """
    with settings(warn_only=True):
        result = sudo('getent passwd {0}'.format(name))
    if result.failed:
        uid = '--uid {0}'.format(uid) if uid else ''
        shell = '--shell {0}'.format(shell) if shell else ''
        sudo('adduser --system --home {0} --group {1} {2} {3}'.format(home, name, uid, shell))


def set_timezone(timezone):
    """
    Set the system timezone to the given value, e.g. 'America/Chicago'.
    """
    sudo('echo {} >/etc/timezone'.format(timezone))
    if is_link('/etc/localtime'):
        sudo('ln --symbolic --force /usr/share/zoneinfo/{} /etc/localtime'.format(timezone))
    else:
        sudo('cp /usr/share/zoneinfo/{} /etc/localtime'.format(timezone))


def agent_sudo(cmd, user=None):
    """
    Run a 'sudo' command on the target server, with full agent forwarding. 
    """
    with settings(forward_agent=True):
        sudo('SSH_AUTH_SOCK=$SSH_AUTH_SOCK {}'.format(cmd), shell=False, user=user)


def put(local_path, remote_path, owner='root:root', **kwargs):
    """
    Put a file on the server, and set its ownership.
    """
    if 'mode' not in kwargs:
        kwargs.setdefault('mirror_local_mode', True)
    kwargs['use_sudo'] = True
    fab_put(local_path, remote_path, **kwargs)
    sudo('chown {0} {1}'.format(owner, remote_path))


def upload_template(local_path, remote_path, owner='root:root', **kwargs):
    """
    Upload a template to the server, and set its ownership.
    """
    if 'mode' not in kwargs:
        kwargs.setdefault('mirror_local_mode', True)
    kwargs['use_sudo'] = True
    fab_upload_template(local_path, remote_path, **kwargs)
    sudo('chown {0} {1}'.format(owner, remote_path))


def upload_mako_template(local_path, remote_path, context={}, **kwargs):
    """
    Render a local file as a Mako template, and upload the result to the server.
    """
    template = Template(filename=local_path)

    temp_dir = tempfile.mkdtemp(prefix='rattail.fablib.')
    temp_path = os.path.join(temp_dir, os.path.basename(local_path))
    with open(temp_path, 'wb') as f:
        f.write(template.render(env=env, **context))
    os.chmod(temp_path, os.stat(local_path).st_mode)

    put(temp_path, remote_path, **kwargs)
    os.remove(temp_path)
    os.rmdir(temp_dir)


class Deployer(object):

    def __init__(self, deploy_path, last_segment='deploy'):
        if not os.path.isdir(deploy_path):
            deploy_path = os.path.abspath(os.path.join(os.path.dirname(deploy_path), last_segment))
        self.deploy_path = deploy_path

    def __call__(self, local_path, remote_path, **kwargs):
        self.deploy(local_path, remote_path, **kwargs)

    def full_path(self, local_path):
        return '{}/{}'.format(self.deploy_path, local_path)

    def local_exists(self, local_path):
        return os.path.exists(self.full_path(local_path))

    def deploy(self, local_path, remote_path, **kwargs):
        local_path = self.full_path(local_path)
        context = kwargs.pop('context', {})
        if local_path.endswith('.template'):
            upload_template(local_path, remote_path, context=env, **kwargs)
        elif local_path.endswith('.mako'):
            upload_mako_template(local_path, remote_path, context=context, **kwargs)
        else:
            put(local_path, remote_path, **kwargs)

    def sudoers(self, local_path, remote_path, mode='0440', **kwargs):
        self.deploy(local_path, '/tmp/sudoers', mode=mode)
        sudo('mv /tmp/sudoers {0}'.format(remote_path))

    def apache_site(self, local_path, name, **kwargs):
        from rattail.fablib.apache import deploy_site
        deploy_site(self, local_path, name, **kwargs)

    def apache_conf(self, local_path, name, **kwargs):
        from rattail.fablib.apache import deploy_conf
        deploy_conf(self, local_path, name, **kwargs)

    def certbot_account(self, uuid, localdir='certbot/account'):
        """
        Deploy files to establish a certbot account on target server
        """
        localdir = localdir.rstrip('/')
        paths = [
            '/etc/letsencrypt/accounts',
            '/etc/letsencrypt/accounts/acme-v01.api.letsencrypt.org',
            '/etc/letsencrypt/accounts/acme-v01.api.letsencrypt.org/directory',
        ]
        final_path = '{}/{}'.format(paths[-1], uuid)
        paths.append(final_path)
        if not exists(final_path):
            mkdir(paths, mode='0700')
            with cd(final_path):
                self.deploy('{}/private_key.json'.format(localdir), 'private_key.json', mode='0600')
                self.deploy('{}/meta.json'.format(localdir), 'meta.json')
                self.deploy('{}/regr.json'.format(localdir), 'regr.json')

    def soffice_daemon(self, local_path, name=None, register=True, start=True, **kwargs):
        if name is None:
            name = local_path.split('/')[-1]
        self.deploy(local_path, '/etc/init.d/{}'.format(name), **kwargs)
        if register:
            sudo('update-rc.d {} defaults'.format(name))
            if start:
                sudo('service {} restart'.format(name))


def make_deploy(deploy_path, last_segment='deploy'):
    """
    Make a ``deploy()`` function, for uploading files to the server.

    During a deployment, one usually needs to upload certain additional files
    to the server.  It's also often necessary to dynamically define certain
    settings etc. within these files.  The :func:`upload_template()` and
    :func:`put()` functions, respectively, handle uploading files which do and
    do not require dynamic variable substitution.

    The return value from ``make_deploy()`` is a function which will call
    ``put()`` or ``upload_template()`` based on whether or not the file path
    ends with ``'.template'``.

    To make the ``deploy()`` function even simpler for the caller, it will
    assume a certain context for local file paths.  This means one only need
    provide a base file name when calling ``deploy()``, and it will be
    interpreted as relative to the function's context path.

    The ``deploy_path`` argument is used to establish the context path for the
    function.  If it is a folder path, it will be used as-is; otherwise it will
    be constructed by joining the parent folder of ``deploy_path`` with the
    value of ``last_segment``.

    Typical usage then is something like::

       from rattail.fablib import make_deploy

       deploy = make_deploy(__file__)

       deploy('rattail/init-filemon', '/etc/init.d/rattail-filemon',
              mode='0755')

       deploy('rattail/rattail.conf.template', '/etc/rattail.conf')

    This shows what is intended to be typical, i.e. where ``__file__`` is the
    only argument required for ``make_deploy()``.  For the above to work will
    require you to have something like this file structure, where
    ``fabfile.py`` is the script which contains the above code::

       myproject/
       |-- fabfile.py
       |-- deploy/
           `-- rattail/
               |-- init-filemon
               |-- rattail.conf.template
    """
    return Deployer(deploy_path, last_segment)


def rsync(host, *paths):
    """
    Runs rsync as root, for the given host and file paths.
    """
    for path in paths:
        assert path.startswith('/')
        path = path.rstrip('/')
        # escape path for rsync
        path = path.replace(' ', r'\\\ ').replace("'", r"\\\'")
        agent_sudo('rsync -aP --del root@{0}:{1}/ {1}'.format(host, path))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV File Utilities

Contains various utilities relating to CSV file processing.

.. note::
   This module is named ``csvutil`` instead of ``csv`` primarily as a
   workaround to the problem of ``PythonService.exe`` insisting on doing
   relative imports.
"""

from __future__ import unicode_literals

import csv
import codecs
from cStringIO import StringIO


class DictWriter(csv.DictWriter):
    """
    Convenience implementation of ``csv.DictWriter``.

    This exists only to provide the :meth:`writeheader()` method on Python 2.6.
    """

    def writeheader(self):
        if hasattr(csv.DictWriter, 'writeheader'):
            return csv.DictWriter.writeheader(self)
        self.writer.writerow(self.fieldnames)


class UTF8Recoder(object):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8.

    .. note::
       This class was stolen from the Python 2.7 documentation.
    """

    def __init__(self, fileobj, encoding, errors='strict'):
        self.errors = errors
        self.reader = codecs.getreader(encoding)(fileobj, errors=self.errors)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf_8')


class UnicodeReader(object):
    """
    A CSV reader which will iterate over lines in a CSV file, which is encoded
    in the given encoding.

    .. note::
       This class was stolen from the Python 2.7 documentation.
    """

    def __init__(self, fileobj, dialect=csv.excel, encoding='utf_8', errors='strict', **kwargs):
        fileobj = UTF8Recoder(fileobj, encoding, errors=errors)
        self.reader = csv.reader(fileobj, dialect=dialect, **kwargs)

    def __iter__(self):
        return self

    def next(self):
        row = self.reader.next()
        return [unicode(x, 'utf_8') for x in row]


class UnicodeDictReader(object):
    """
    A CSV Dict reader which will iterate over lines in a CSV file, which is
    encoded in the given encoding.
    """

    def __init__(self, fileobj, dialect=csv.excel, encoding='utf_8', errors='strict', **kwargs):
        fileobj = UTF8Recoder(fileobj, encoding, errors=errors)
        self.reader = csv.reader(fileobj, dialect=dialect, **kwargs)
        self.header = self.reader.next()

    def next(self):
        row = self.reader.next()
        vals = [unicode(s, 'utf_8') for s in row]
        return dict((self.header[i], vals[i]) for i in range(len(self.header)))

    def __iter__(self):
        return self


class UnicodeWriter(object):
    """
    A CSV writer which will write rows to CSV file "f", which is encoded in the
    given encoding.

    .. note::
       This class was stolen from the Python 2.7 documentation.
    """

    def __init__(self, f, dialect='excel', encoding='utf_8', encoding_errors='strict', **kwargs):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwargs)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)(encoding_errors)

    def writerow(self, row):
        self.writer.writerow([s.encode('utf_8') for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode('utf_8')
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class UnicodeDictWriter(UnicodeWriter):
    """
    A ``DictWriter``-ish class which accepts row data as Unicode and can write
    to the file with any encoding.

    .. note::
       This logic was stolen from a `Django snippet`_.  The original docstring
       from this snippet follows ("sic" applies here; our logic uses 'utf_8'
       encoding and regular 'excel' dialect by default):

    A CSV writer that produces Excel-compatibly CSV files from unicode data.
    Uses UTF-16 and tabs as delimeters - it turns out this is the only way to
    get unicode data in to Excel using CSV.

    Usage example::

       fp = open('my-file.csv', 'wb')
       writer = UnicodeDictWriter(fp, ['name', 'age', 'shoesize'])
       writer.writerows([
           {'name': u'Bob', 'age': 22, 'shoesize': 7},
           {'name': u'Sue', 'age': 28, 'shoesize': 6},
           {'name': u'Ben', 'age': 31, 'shoesize': 8},
           # \xc3\x80 is LATIN CAPITAL LETTER A WITH MACRON
           {'name': '\xc4\x80dam'.decode('utf8'), 'age': 11, 'shoesize': 4},
       ])
       fp.close()

    Initially derived from http://docs.python.org/lib/csv-examples.html

    .. _`Django snippet`: https://djangosnippets.org/snippets/993/
    """
    
    def __init__(self, f, fields, dialect='excel', encoding='utf_8', **kwds):
        super(UnicodeDictWriter, self).__init__(f, dialect, encoding, **kwds)
        self.fields = fields
    
    def writerow(self, drow):
        row = [drow.get(field, '') for field in self.fields]
        super(UnicodeDictWriter, self).writerow(row)

    def writeheader(self):
        super(UnicodeDictWriter, self).writerow(self.fields)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from rattail.db import model
from rattail.tests import DataTestCase


class TestModelBase(DataTestCase):

    def test_repr(self):
        # We test this via the Setting class for convenience, since it doesn't
        # have a proper `repr()` method of its own.  I started to add an
        # assertion to prove that, but then wasn't sure how to go about it.
        setting = model.Setting()
        self.assertEqual(repr(setting), b"Setting(name=None)")
        setting.name = 'rattail.test.repr'
        self.assertEqual(repr(setting), b"Setting(name=u'rattail.test.repr')")


class TestSetting(DataTestCase):

    def test_unicode(self):
        setting = model.Setting()
        self.assertEqual(unicode(setting), "")
        setting.name = 'rattail.test.unicode'
        self.assertEqual(unicode(setting), "rattail.test.unicode")


class TestChange(DataTestCase):

    def test_repr(self):
        change = model.Change()
        self.assertEqual(repr(change), b"Change(class_name=None, instance_uuid=None, deleted=None)")
        change.class_name = 'Widget'
        self.assertEqual(repr(change), b"Change(class_name=u'Widget', instance_uuid=None, deleted=None)")
        change.object_key = '42'
        self.assertEqual(repr(change), b"Change(class_name=u'Widget', instance_uuid=u'42', deleted=None)")
        change.deleted = False
        self.assertEqual(repr(change), b"Change(class_name=u'Widget', instance_uuid=u'42', deleted=False)")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric Library for PostgreSQL
"""

from __future__ import unicode_literals, absolute_import

import os
import re

from fabric.api import sudo, run, get, hide, abort, put, local

from rattail.fablib import apt


def install():
    """
    Install the PostgreSQL database service
    """
    apt.install('postgresql')


def get_version():
    """
    Fetch the version of PostgreSQL running on the target system
    """
    result = sudo('psql --version')
    if result.succeeded:
        match = re.match(r'^psql \(PostgreSQL\) (\d+\.\d+)\.\d+', result)
        if match:
            return float(match.group(1))


def sql(sql, database=''):
    """
    Execute some SQL as the 'postgres' user.
    """
    return sudo('sudo -u postgres psql --tuples-only --no-align --command="{0}" {1}'.format(sql, database), shell=False)


def script(path, database='', user=None, password=None):
    """
    Execute a SQL script.  By default this will run as 'postgres' user, but can
    use PGPASSWORD authentication if necessary.
    """
    if user and password:
        with hide('running'):
            return sudo(" PGPASSWORD='{}' psql --host=localhost --username='{}' --file='{}' {}".format(
                password, user, path, database))

    else: # run as postgres
        return sudo("sudo -u postgres psql --file='{}' {}".format(path, database), shell=False)


def user_exists(name):
    """
    Determine if a given PostgreSQL user exists.
    """
    user = sql("SELECT rolname FROM pg_roles WHERE rolname = '{0}'".format(name))
    return bool(user)


def create_user(name, password=None, checkfirst=True, createdb=False):
    """
    Create a PostgreSQL user account.
    """
    if not checkfirst or not user_exists(name):
        createdb = '--{}createdb'.format('' if createdb else 'no-')
        sudo('sudo -u postgres createuser {createdb} --no-createrole --no-superuser {name}'.format(
            createdb=createdb, name=name))
        if password:
            set_user_password(name, password)


def set_user_password(name, password):
    """
    Set the password for a PostgreSQL user account
    """
    with hide('running'):
        sql("ALTER USER \\\"{}\\\" PASSWORD '{}';".format(name, password))


def db_exists(name):
    """
    Determine if a given PostgreSQL database exists.
    """
    db = sql("SELECT datname FROM pg_database WHERE datname = '{0}'".format(name))
    return db == name


def create_db(name, owner=None, checkfirst=True):
    """
    Create a PostgreSQL database.
    """
    if not checkfirst or not db_exists(name):
        args = '--owner={0}'.format(owner) if owner else ''
        sudo('sudo -u postgres createdb {0} {1}'.format(args, name), shell=False)


def drop_db(name, checkfirst=True):
    """
    Drop a PostgreSQL database.
    """
    if not checkfirst or db_exists(name):
        sudo('sudo -u postgres dropdb {0}'.format(name), shell=False)


def download_db(name, destination=None):
    """
    Download a database from the "current" server.
    """
    if destination is None:
        destination = './{0}.sql.gz'.format(name)
    run('touch {0}.sql'.format(name))
    run('chmod 0666 {0}.sql'.format(name))
    sudo('sudo -u postgres pg_dump --file={0}.sql {0}'.format(name), shell=False)
    run('gzip --force {0}.sql'.format(name))
    get('{0}.sql.gz'.format(name), destination)
    run('rm {0}.sql.gz'.format(name))


def clone_db(name, owner, download, user='rattail', force=False, workdir=None):
    """
    Clone a database from a (presumably live) server

    :param name: Name of the database.

    :param owner: Username of the user who is to own the database.

    :param force: Whether the target database should be forcibly dropped, if it
       exists already.
    """
    if db_exists(name):
       if force:
           drop_db(name, checkfirst=False)
       else:
           abort("Database '{}' already exists!".format(name))

    create_db(name, owner=owner, checkfirst=False)

    # upload database dump to target server
    if workdir:
        curdir = os.getcwd()
        os.chdir(workdir)
    download('{}.sql.gz'.format(name), user=user)
    put('{}.sql.gz'.format(name))
    local('rm {}.sql.gz'.format(name))
    if workdir:
        os.chdir(curdir)

    # restore database on target server
    run('gunzip --force {}.sql.gz'.format(name))
    sudo('sudo -u postgres psql --file={0}.sql {0}'.format(name), shell=False)
    run('rm {}.sql'.format(name))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for purchase order batches
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy import orm

from rattail.db import model, api
from rattail.batch import BatchHandler
from rattail.time import make_utc


class PurchaseBatchHandler(BatchHandler):
    """
    Handler for purchase order batches.
    """
    batch_model_class = model.PurchaseBatch

    # set this to True for handler to skip various "case" logic
    ignore_cases = False

    def should_populate(self, batch):
        # TODO: this probably should change soon, for now this works..
        return batch.purchase and batch.mode in (self.enum.PURCHASE_BATCH_MODE_RECEIVING,
                                                 self.enum.PURCHASE_BATCH_MODE_COSTING)

    def populate(self, batch, progress=None):
        assert batch.purchase and batch.mode in (self.enum.PURCHASE_BATCH_MODE_RECEIVING,
                                                 self.enum.PURCHASE_BATCH_MODE_COSTING)

        def append(item, i):
            row = model.PurchaseBatchRow()
            row.item = item
            row.product = item.product
            row.cases_ordered = item.cases_ordered
            row.units_ordered = item.units_ordered
            row.cases_received = item.cases_received
            row.units_received = item.units_received
            row.po_unit_cost = item.po_unit_cost
            row.po_total = item.po_total
            if batch.mode == self.enum.PURCHASE_BATCH_MODE_COSTING:
                row.invoice_unit_cost = item.invoice_unit_cost
                row.invoice_total = item.invoice_total
            batch.add_row(row)
            self.refresh_row(row)

        assert self.progress_loop(append, batch.purchase.items, progress,
                                  message="Adding initial rows to batch")

    def refresh(self, batch, progress=None):
        if batch.mode == self.enum.PURCHASE_BATCH_MODE_ORDERING:
            batch.po_total = 0
        elif batch.mode in (self.enum.PURCHASE_BATCH_MODE_RECEIVING,
                            self.enum.PURCHASE_BATCH_MODE_COSTING):
            batch.invoice_total = 0
        return super(PurchaseBatchHandler, self).refresh(batch, progress=progress)

    def refresh_batch_status(self, batch):
        if any([not row.product_uuid for row in batch.active_rows()]):
            batch.status_code = batch.STATUS_UNKNOWN_PRODUCT
        else:
            batch.status_code = batch.STATUS_OK

    def refresh_row(self, row, initial=False):
        """
        Refreshing a row will A) assume that ``row.product`` is already set to
        a valid product, and B) update various other fields on the row
        (description, size, etc.)  to reflect the current product data.  It
        also will adjust the batch PO total per the row PO total.
        """
        batch = row.batch
        product = row.product
        if not product:
            if row.upc:
                session = orm.object_session(row)
                product = api.get_product_by_upc(session, row.upc)
            if not product:
                # TODO: should we do more stuff here..?
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return
            row.product = product

        cost = row.product.cost_for_vendor(batch.vendor) or row.product.cost
        row.upc = product.upc
        row.item_id = product.item_id
        row.brand_name = six.text_type(product.brand or '')
        row.description = product.description
        row.size = product.size
        if product.department:
            row.department_number = product.department.number
            row.department_name = product.department.name
        else:
            row.department_number = None
            row.department_name = None
        row.vendor_code = cost.code if cost else None
        new_case_quantity = cost.case_size if cost else None
        if new_case_quantity:
            row.case_quantity = new_case_quantity

        self.refresh_totals(row, cost, initial)

        if batch.mode == self.enum.PURCHASE_BATCH_MODE_ORDERING:
            row.status_code = row.STATUS_OK

        elif batch.mode in (self.enum.PURCHASE_BATCH_MODE_RECEIVING,
                            self.enum.PURCHASE_BATCH_MODE_COSTING):
            if (row.cases_received is None and row.units_received is None and
                row.cases_damaged is None and row.units_damaged is None and
                row.cases_expired is None and row.units_expired is None and
                row.cases_mispick is None and row.units_mispick is None):
                row.status_code = row.STATUS_INCOMPLETE
            else:
                if self.get_units_ordered(row) != self.get_units_accounted_for(row):
                    row.status_code = row.STATUS_ORDERED_RECEIVED_DIFFER
                else:
                    row.status_code = row.STATUS_OK

    def refresh_totals(self, row, cost, initial):
        batch = row.batch

        if batch.mode == self.enum.PURCHASE_BATCH_MODE_ORDERING:
            row.po_unit_cost = self.get_unit_cost(row.product, batch.vendor)
            if row.po_unit_cost:
                row.po_total = row.po_unit_cost * self.get_units_ordered(row)
                batch.po_total = (batch.po_total or 0) + row.po_total
            else:
                row.po_total = None

        elif batch.mode in (self.enum.PURCHASE_BATCH_MODE_RECEIVING,
                            self.enum.PURCHASE_BATCH_MODE_COSTING):
            row.invoice_unit_cost = (cost.unit_cost if cost else None) or row.po_unit_cost
            if row.invoice_unit_cost:
                row.invoice_total = row.invoice_unit_cost * self.get_units_accounted_for(row)
                batch.invoice_total = (batch.invoice_total or 0) + row.invoice_total
            else:
                row.invoice_total = None

    def get_unit_cost(self, product, vendor):
        """
        Must return the PO unit cost for the given product, from the given vendor.
        """
        cost = product.cost_for_vendor(vendor) or product.cost
        if cost:
            return cost.unit_cost

    def get_units_ordered(self, row):
        return (row.units_ordered or 0) + (row.case_quantity or 1) * (row.cases_ordered or 0)

    def get_units_received(self, row):
        return (row.units_received or 0) + row.case_quantity * (row.cases_received or 0)

    def get_units_shipped(self, row):
        units_damaged = (row.units_damaged or 0) + row.case_quantity * (row.cases_damaged or 0)
        units_expired = (row.units_expired or 0) + row.case_quantity * (row.cases_expired or 0)
        return self.get_units_received(row) + units_damaged + units_expired

    def get_units_accounted_for(self, row):
        units_mispick = (row.units_mispick or 0) + row.case_quantity * (row.cases_mispick or 0)
        return self.get_units_shipped(row) + units_mispick

    def update_order_counts(self, purchase, progress=None):

        def update(item, i):
            if item.product:
                inventory = item.product.inventory or model.ProductInventory(product=item.product)
                inventory.on_order = (inventory.on_order or 0) + (item.units_ordered or 0) + (
                    (item.cases_ordered or 0) * (item.case_quantity or 1))

        self.progress_loop(update, purchase.items, progress,
                           message="Updating inventory counts")

    def execute(self, batch, user, progress=None):
        """
        Default behavior for executing a purchase batch will create a new
        purchase, by invoking :meth:`make_purchase()`.
        """
        session = orm.object_session(batch)

        if batch.mode == self.enum.PURCHASE_BATCH_MODE_ORDERING:
            purchase = self.make_purchase(batch, user, progress=progress)
            self.update_order_counts(purchase, progress=progress)
            return purchase

        elif batch.mode == self.enum.PURCHASE_BATCH_MODE_RECEIVING:
            with session.no_autoflush:
                return self.receive_purchase(batch, progress=progress)

        elif batch.mode == self.enum.PURCHASE_BATCH_MODE_COSTING:
            # TODO: finish this...
            # with session.no_autoflush:
            #     return self.cost_purchase(batch, progress=progress)
            purchase = batch.purchase
            purchase.invoice_date = batch.invoice_date
            purchase.status = self.enum.PURCHASE_STATUS_COSTED
            return purchase

        assert False

    def make_credits(self, batch, progress=None):
        session = orm.object_session(batch)
        mapper = orm.class_mapper(model.PurchaseBatchCredit)

        def copy(row, i):
            for batch_credit in row.credits:
                credit = model.PurchaseCredit()
                for prop in mapper.iterate_properties:
                    if isinstance(prop, orm.ColumnProperty) and hasattr(credit, prop.key):
                        setattr(credit, prop.key, getattr(batch_credit, prop.key))
                credit.status = self.enum.PURCHASE_CREDIT_STATUS_NEW
                session.add(credit)

        return self.progress_loop(copy, batch.active_rows(), progress,
                                  message="Creating purchase credits")

    def make_purchase(self, batch, user, progress=None):
        """
        Effectively clones the given batch, creating a new Purchase in the
        Rattail system.
        """
        session = orm.object_session(batch)
        purchase = model.Purchase()

        # TODO: should be smarter and only copy certain fields here
        skip_fields = [
            'date_received',
        ]
        for prop in orm.object_mapper(batch).iterate_properties:
            if prop.key in skip_fields:
                continue
            if hasattr(purchase, prop.key):
                setattr(purchase, prop.key, getattr(batch, prop.key))

        def clone(row, i):
            item = model.PurchaseItem()
            # TODO: should be smarter and only copy certain fields here
            for prop in orm.object_mapper(row).iterate_properties:
                if hasattr(item, prop.key):
                    setattr(item, prop.key, getattr(row, prop.key))
            purchase.items.append(item)

        assert self.progress_loop(clone, batch.active_rows(), progress,
                                  message="Creating purchase items")

        purchase.created = make_utc()
        purchase.created_by = user
        purchase.status = self.enum.PURCHASE_STATUS_ORDERED
        session.add(purchase)
        batch.purchase = purchase
        return purchase

    def receive_purchase(self, batch, progress=None):
        """
        Update the purchase for the given batch, to indicate received status.
        """
        session = orm.object_session(batch)
        purchase = batch.purchase
        if not purchase:
            batch.purchase = purchase = model.Purchase()

            # TODO: should be smarter and only copy certain fields here
            skip_fields = [
                'date_received',
            ]
            for prop in orm.object_mapper(batch).iterate_properties:
                if prop.key in skip_fields:
                    continue
                if hasattr(purchase, prop.key):
                    setattr(purchase, prop.key, getattr(batch, prop.key))

        purchase.invoice_number = batch.invoice_number
        purchase.invoice_date = batch.invoice_date
        purchase.invoice_total = batch.invoice_total
        purchase.date_received = batch.date_received

        # determine which fields we'll copy when creating new purchase item
        copy_fields = []
        for prop in orm.class_mapper(model.PurchaseItem).iterate_properties:
            if hasattr(model.PurchaseBatchRow, prop.key):
                copy_fields.append(prop.key)

        def update(row, i):
            item = row.item
            if not item:
                row.item = item = model.PurchaseItem()
                for field in copy_fields:
                    setattr(item, field, getattr(row, field))
                purchase.items.append(item)

            item.cases_received = row.cases_received
            item.units_received = row.units_received
            item.cases_damaged = row.cases_damaged
            item.units_damaged = row.units_damaged
            item.cases_expired = row.cases_expired
            item.units_expired = row.units_expired
            item.invoice_line_number = row.invoice_line_number
            item.invoice_case_cost = row.invoice_case_cost
            item.invoice_unit_cost = row.invoice_unit_cost
            item.invoice_total = row.invoice_total

        with session.no_autoflush:
            assert self.progress_loop(update, batch.active_rows(), progress,
                                      message="Updating purchase line items")

        purchase.status = self.enum.PURCHASE_STATUS_RECEIVED
        return purchase



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
from unittest import TestCase

import lockfile
from fixture import TempIO

from rattail import files


class TestLockingCopy(TestCase):

    def setUp(self):
        self.tmp = TempIO()
        self.src = self.tmp.mkdir(u'src')
        self.dst = self.tmp.mkdir(u'dst')
        self.src_file = self.src.putfile(u'somefile', b'')

    def test_normal_copy_succeeds(self):
        files.locking_copy(self.src_file, self.dst)
        dst_file = os.path.join(self.dst, u'somefile')
        self.assertTrue(os.path.exists(dst_file))
        self.assertTrue(os.path.isfile(dst_file))
        self.assertFalse(os.path.exists(os.path.join(self.dst, u'somefile.lock')))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric library for CORE-POS (IS4C)
"""

from __future__ import unicode_literals, absolute_import

import webbrowser

from fabric.api import sudo, cd, env, run
from fabric.contrib.files import exists

from rattail.fablib import apt, mysql, mkdir


def install_fannie(rootdir, branch='version-2.3', first_time=None, url=None):
    """
    Install the Fannie app to the given location
    """
    if first_time is None:
        first_time = not exists(rootdir)
    mkdir(rootdir)
    with cd(rootdir):

        # fannie source
        if not exists('IS4C'):
            sudo('git clone https://github.com/CORE-POS/IS4C.git')
            with cd('IS4C'):
                sudo('git checkout {}'.format(branch))
        with cd('IS4C'):
            sudo('git pull')

        # composer
        if not exists('composer.phar'):
            sudo('curl -sS https://getcomposer.org/installer | php')
        mkdir('/var/www/.composer', owner='www-data')

        # fannie dependencies
        with cd('IS4C'):
            mkdir(['vendor', 'fannie/src/javascript/composer-components'], owner='www-data')
            sudo('../composer.phar install', user='www-data')

        # shadowread
        with cd('IS4C/fannie/auth/shadowread'):
            sudo('make')
            sudo('make install')

        # fannie config
        with cd('IS4C/fannie'):
            if not exists('config.php'):
                sudo('cp config.php.dist config.php')
                sudo('chown root:www-data config.php')
                sudo('chmod 0660 config.php')

        # fannie logging
        with cd('IS4C/fannie/logs'):
            for name in ['fannie', 'debug_fannie', 'queries', 'php-errors', 'dayend']:
                sudo('touch {}.log'.format(name))
                sudo('chown www-data:www-data {}.log'.format(name))

    # fannie databases
    mysql.create_user('is4c', host='%', password='is4c')
    mysql.create_db('core_op', user="is4c@'%'")
    mysql.create_db('core_trans', user="is4c@'%'")
    mysql.create_db('trans_archive', user="is4c@'%'")

    # fannie web installer
    if first_time:
        url = url or getattr(env, 'fannie_rooturl', 'http://localhost/fannie/')
        webbrowser.open_new_tab('{}/install/'.format(url.rstrip('/')))



# -*- coding: utf-8 -*-
"""make cost.case_size decimal

Revision ID: c384c65995db
Revises: 8c17754a6bbf
Create Date: 2016-11-11 19:20:25.017965

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'c384c65995db'
down_revision = u'8c17754a6bbf'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # product_cost
    op.alter_column('product_cost', 'case_size', type_=sa.Numeric(precision=9, scale=4))


def downgrade():

    # product_cost
    op.alter_column('product_cost', 'case_size', type_=sa.Integer())



# -*- coding: utf-8 -*-
"""add batch_pricing_row.vendor

Revision ID: 3dc08453a5d3
Revises: 45d7d92f4aa6
Create Date: 2017-02-02 15:26:15.851844

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3dc08453a5d3'
down_revision = u'45d7d92f4aa6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # batch_pricing_row
    op.add_column('batch_pricing_row', sa.Column('vendor_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'batch_pricing_row_fk_vendor', 'batch_pricing_row', 'vendor', ['vendor_uuid'], ['uuid'])


def downgrade():

    # batch_pricing_row
    op.drop_constraint(u'batch_pricing_row_fk_vendor', 'batch_pricing_row', type_='foreignkey')
    op.drop_column('batch_pricing_row', 'vendor_uuid')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric library for the APT package system
"""

from __future__ import unicode_literals, absolute_import

from fabric.api import sudo, settings
from fabric.contrib.files import append

from rattail.fablib import make_deploy, get_debian_version, get_ubuntu_version


deploy = make_deploy(__file__)


def install(*packages, **kwargs):
    """
    Install one or more packages via ``apt-get install``.
    """
    frontend = kwargs.get('frontend', 'noninteractive')
    target = kwargs.get('target_release')
    target = '--target-release={}'.format(target) if target else ''
    force_yes = ' --force-yes' if kwargs.get('force_yes') else ''
    sudo('DEBIAN_FRONTEND={} apt-get --assume-yes {}{} install {}'.format(
            frontend, target, force_yes, ' '.join(packages)))


def purge(*packages):
    """
    Uninstall and purge config for given packages
    """
    sudo('apt-get --assume-yes purge {}'.format(' '.join(packages)))


def update():
    """
    Perform an ``apt-get update`` operation.
    """
    sudo('apt-get update')


def add_repository(repo):
    """
    Add a new APT repository
    """
    sudo('add-apt-repository --yes {}'.format(repo))
    update()


def add_source(entry):
    """
    Add a new entry to the apt/sources.list file
    """
    append('/etc/apt/sources.list', entry, use_sudo=True)
    update()


def dist_upgrade(frontend='noninteractive'):
    """
    Perform a full ``apt-get dist-upgrade`` operation.
    """
    update()
    sudo('DEBIAN_FRONTEND={} apt-get --assume-yes dist-upgrade'.format(frontend))


def configure_listchanges():
    """
    Configure apt listchanges to never use a frontend.
    """
    deploy('apt/listchanges.conf', '/etc/apt/listchanges.conf')


def install_emacs():
    """
    Install the Emacs editor
    """
    with settings(warn_only=True):
        result = sudo('which emacs')
    if result.succeeded:
        return

    emacs = 'emacs-nox'
    debian_version = get_debian_version()
    if debian_version:
        if debian_version < 8:
            emacs = 'emacs23-nox'
    else:
        ubuntu_version = get_ubuntu_version()
        if ubuntu_version and ubuntu_version < 16:
            emacs = 'emacs23-nox'

    install(emacs, 'emacs-goodies-el')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric Library for SSH
"""

from __future__ import unicode_literals, absolute_import

import warnings

from fabric.api import sudo, cd, settings
from fabric.contrib.files import exists, sed, append

from rattail.fablib import mkdir, agent_sudo
from rattail.fablib.python import cdvirtualenv


def cache_host_key(host, for_user='root', with_agent=False, warn_only=True, identity=''):
    """
    Cache the SSH host key for the given host, for the given user.
    """
    user = None if for_user == 'root' else for_user
    _sudo = agent_sudo if with_agent else sudo
    if identity:
        identity = '-i {}'.format(identity)
    cmd = 'ssh {} -o StrictHostKeyChecking=no {} echo'.format(identity, host)
    if warn_only:
        with settings(warn_only=True):
            _sudo(cmd, user=user)
    else:
        _sudo(cmd, user=user)


def uncache_host_key(host, for_user='root'):
    """
    Remove the cached SSH host key for the given host, for the given user.
    """
    user = None if for_user == 'root' else for_user
    sudo('ssh-keygen -R {}'.format(host), user=user)


def restart():
    """
    Restart the OpenSSH service
    """
    sudo('service ssh restart')


def configure(allow_root=False):
    """
    Configure the OpenSSH service
    """
    path = '/etc/ssh/sshd_config'
    value = 'without-password' if allow_root else 'no'
    sed(path, r'^PermitRootLogin\s+.*', 'PermitRootLogin {}'.format(value), use_sudo=True)
    entry = 'PasswordAuthentication no'
    sed(path, r'^PasswordAuthentication\s+.*', entry, use_sudo=True)
    append(path, entry, use_sudo=True)
    restart()


def configure_ssh(restrict_root=True):
    warnings.warn("Function `ssh.configure_ssh()` is deprecated, please "
                  "use `ssh.configure()` instead.", DeprecationWarning)
    return configure(allow_root=not restrict_root)


def establish_identity(envname, comment, user='rattail', home='/var/lib/rattail'):
    """
    Generate a SSH key pair and configure it for local use.
    """
    home = home.rstrip('/')
    sshdir = '{0}/.ssh'.format(home)
    owner='{0}:{0}'.format(user)
    mkdir(sshdir, owner=owner, mode='0700')
    with cd(sshdir):
        if not exists('authorized_keys'):
            sudo('touch authorized_keys')
        sudo('chown {0} authorized_keys'.format(owner))
        sudo('chmod 0600 authorized_keys')
    with cdvirtualenv(envname, 'app'):
        mkdir('ssh', owner=owner, mode='0700')
    with cdvirtualenv(envname, 'app/ssh'):
        if not exists('id_rsa', use_sudo=True):
            sudo("ssh-keygen -C '{0}' -P '' -f id_rsa".format(comment))
            sudo('cat id_rsa.pub >> {0}/authorized_keys'.format(sshdir))
        sudo('chown {0} id_rsa*'.format(owner))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for Dutch Valley
"""

from __future__ import unicode_literals, absolute_import

import re
import datetime
import logging
from decimal import Decimal

import xlrd

from rattail.db import model
from rattail.vendors.catalogs import CatalogParser


log = logging.getLogger(__name__)


class DutchValleyCatalogParser(CatalogParser):
    """
    Vendor catalog parser for Dutch Valley.
    """
    key = 'rattail.contrib.dutchvalley'
    display = "Dutch Valley"
    vendor_key = 'dutch-valley'

    def parse_effective_date(self, path):
        book = xlrd.open_workbook(path)
        sheet = book.sheet_by_index(0)
        date = sheet.cell_value(0, 0)
        date = xlrd.xldate_as_tuple(date, book.datemode)
        return datetime.datetime(*date).date()

    def parse_rows(self, path, progress=None):
        book = xlrd.open_workbook(path)
        sheet = book.sheet_by_index(0)
        code_pattern = re.compile(r'^\d{3} \d{3}$')
        description_pattern = re.compile(r'^(\d+(?:\.\d+)?)((?:/(\d+(?:\.\d+)?))?\w+) (.*)$')
        for r in range(sheet.nrows):

            # Product code is in first column...but so is a lot of other
            # garbage.  Use strict regex to find good rows.
            code = sheet.cell_value(r, 0)
            if not isinstance(code, basestring):
                continue
            if not code_pattern.match(code):
                continue
            code = code.replace(' ', '')
            
            row = model.VendorCatalogRow()
            row.vendor_code = code

            # Try to parse unit and/or case size from description.
            description = sheet.cell_value(r, 1)
            match = description_pattern.match(description)
            if match:
                row.description = match.group(3) or ''
                row.size = match.group(2)
                row.case_size = int(float(match.group(1)))
            else:
                log.warning("description doesn't match pattern at row {0}: {1}".format(
                        r + 1, repr(description)))
                row.description = description or ''

            case_cost = sheet.cell_value(r, 2)
            try:
                row.case_cost = Decimal(unicode(case_cost))
            except ValueError:
                log.warning("bad case cost at row {0}: {1}".format(r + 1, repr(case_cost)))

            unit_cost = sheet.cell_value(r, 4)
            try:
                row.unit_cost = Decimal(unicode(unit_cost))
            except ValueError:
                log.warning("bad unit cost at row {0}: {1}".format(r + 1, repr(unit_cost)))

            yield row



# -*- coding: utf-8 -*-
"""add user events

Revision ID: 1242c7a0e24a
Revises: cc3de1ca0689
Create Date: 2017-08-04 16:41:40.837455

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '1242c7a0e24a'
down_revision = u'cc3de1ca0689'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # user_event
    op.create_table('user_event',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('type_code', sa.Integer(), nullable=False),
                    sa.Column('occurred', sa.DateTime(), nullable=True),
                    sa.ForeignKeyConstraint(['user_uuid'], [u'user.uuid'], name=u'user_x_role_fk_user'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # user_event
    op.drop_table('user_event')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Catalog batches
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler
from rattail.vendors.catalogs import require_catalog_parser


class VendorCatalogHandler(BatchHandler):
    """
    Handler for vendor catalog batches.
    """
    batch_model_class = model.VendorCatalog

    def should_populate(self, batch):
        # all vendor catalogs must come from data file
        return True

    def setup(self, batch, progress=None):
        self.vendor = batch.vendor
        self.products = {'upc': {}, 'vendor_code': {}}
        session = orm.object_session(batch)
        products = session.query(model.Product)\
                          .options(orm.joinedload(model.Product.brand))\
                          .options(orm.joinedload(model.Product.costs))

        def cache(product, i):
            if product.upc:
                self.products['upc'][product.upc] = product
            cost = product.cost_for_vendor(self.vendor)
            product.vendor_cost = cost
            if cost and cost.code:
                self.products['vendor_code'][cost.code] = product

        assert self.progress_loop(cache, products.all(), progress,
                                  message="Caching products by UPC and vendor code")

    setup_populate = setup
    setup_refresh = setup

    def populate(self, batch, progress=None):
        """
        Pre-fill batch with row data from an input data file, leveraging a
        specific catalog parser.
        """
        assert batch.filename and batch.parser_key
        session = orm.object_session(batch)
        path = batch.filepath(self.config)
        parser = require_catalog_parser(batch.parser_key)
        parser.session = session
        parser.vendor = batch.vendor
        batch.effective = parser.parse_effective_date(path)
        self.setup_populate(batch, progress=progress)

        def append(row, i):
            self.add_row(batch, row)
            if i % 5000 == 0:
                session.flush()

        data = list(parser.parse_rows(path, progress=progress))
        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def identify_product(self, row):
        product = None
        if row.upc:
            product = self.products['upc'].get(row.upc)
        if not product and row.vendor_code:
            product = self.products['vendor_code'].get(row.vendor_code)
        return product

    def refresh_row(self, row):
        """
        Inspect a single row from a catalog, and set its attributes based on
        whether or not the product exists, if we already have a cost record for
        the vendor, if the catalog contains a change etc.  Note that the
        product lookup is done first by UPC and then by vendor item code.
        """
        row.product = self.identify_product(row)
        if not row.product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        row.upc = row.product.upc
        row.brand_name = row.product.brand.name if row.product.brand else None
        row.description = row.product.description
        row.size = row.product.size

        old_cost = row.product.vendor_cost
        if not old_cost:
            row.status_code = row.STATUS_NEW_COST
            return

        row.old_vendor_code = old_cost.code
        row.old_case_size = old_cost.case_size

        row.old_case_cost = old_cost.case_cost
        if row.case_cost is not None and row.old_case_cost is not None:
            row.case_cost_diff = row.case_cost - row.old_case_cost

        row.old_unit_cost = old_cost.unit_cost
        if row.unit_cost is not None and row.old_unit_cost is not None:
            row.unit_cost_diff = row.unit_cost - row.old_unit_cost

        diff = self.cost_differs(row, old_cost)
        if diff:
            row.status_code = row.STATUS_UPDATE_COST
            row.status_text = diff
        else:
            row.status_code = row.STATUS_NO_CHANGE

    def cost_differs(self, row, cost):
        """
        Compare a batch row with a cost record to determine whether they match
        or differ.
        """
        if row.vendor_code is not None and row.vendor_code != cost.code:
            return "new vendor code {} differs from old code {}".format(
                repr(row.vendor_code), repr(cost.code))
        if row.case_cost is not None and row.case_cost != cost.case_cost:
            return "new case cost {} differs from old cost {}".format(
                row.case_cost, cost.case_cost)
        if row.unit_cost is not None and row.unit_cost != cost.unit_cost:
            return "new unit cost {} differs from old cost {}".format(
                row.unit_cost, cost.unit_cost)
